#include "launchelf.h"


int checkVCDstuff(char *path, int mode)
{
	int fd, size, ret;
	char fullpath[MAX_PATH], tmp[MAX_PATH], *p;
	unsigned char datacheck[32] __attribute((aligned(16)));
	iox_stat_t stat;
	char arg0[MAX_PATH];
	char *exec_args[1] = { arg0 };

	/* File extension check */
	size = strlen(path);
//	if(strncmp(path + size-4, ".VCD", 4) != 0) return -12; // The file has no .VCD extension
	if(path[size-4] != 0x2E || path[size-3] != 0x56 || path[size-2] != 0x43 || path[size-1] != 0x44) return -12; // The file has no .VCD extension

	strcpy(fullpath, path);

	/* VCD storage check */
	if(!strncmp(fullpath, "hdd0:", 5)) {
		p = &path[5];
		if(*p == '/') p++;
		sprintf(tmp, "hdd0:%s", p);
		p = strchr(tmp, '/');
		sprintf(fullpath, "pfs0:%s", p);
		*p = 0;

		if(!strncmp(tmp, "hdd0:__.", 8) || !strncmp(tmp, "hdd0:PP.", 8) || !strncmp(tmp, "hdd0:__.POPS", 12) || !strncmp(tmp, "hdd0:__.POPS0", 13) || !strncmp(tmp, "hdd0:__.POPS1", 13) || !strncmp(tmp, "hdd0:__.POPS2", 13) || !strncmp(tmp, "hdd0:__.POPS3", 13) || !strncmp(tmp, "hdd0:__.POPS4", 13) || !strncmp(tmp, "hdd0:__.POPS5", 13) || !strncmp(tmp, "hdd0:__.POPS6", 13) || !strncmp(tmp, "hdd0:__.POPS7", 13) || !strncmp(tmp, "hdd0:__.POPS8", 13) || !strncmp(tmp, "hdd0:__.POPS9", 13)) {
			if(strncmp(tmp, "hdd0:__.POPS", 12) != 0 && strncmp(fullpath, "pfs0:/IMAGE0.VCD", 16) != 0) return -10;
			if((ret = mountParty(tmp)) < 0) return -13; // Can't mount the VCD partition
			fullpath[3] += ret;
		} else {
			return -10;  // VCD path isn't POPStarter supported (illegal partition)
		}
	} else if(!strncmp(fullpath, "mass", 4)){
		char *pathSep;

		pathSep = strchr(path, '/');
		if(pathSep && (pathSep-path<7) && pathSep[-1]==':') strcpy(fullpath+(pathSep-path), pathSep+1);
		if(strncmp(fullpath, "mass:POPS/", 10) != 0) return -10;
	} else {
		return -10;  // VCD path isn't POPStarter supported (illegal device)
	}

	/* VCD validity check */
	if((fd = genOpen(fullpath, O_RDONLY)) < 0) return 0; // Can't open the VCD for reading
	size = genLseek(fd, 0, SEEK_END);
	if(size < 1102672) {
		genClose(fd);
		return -11; // File is too small to be a valid VCD
	}
	genLseek(fd, 0, SEEK_SET);
	genRead(fd, datacheck, 4);
	if(datacheck[0] != 0x41 || datacheck[1] != 0x00 || datacheck[2] != 0xA0 || datacheck[3] != 0x00) {
		genClose(fd);
		return -11; // VCD has invalid header data (descriptor)
	}
	genLseek(fd, 1024, SEEK_SET);
	genRead(fd, datacheck, 16);
	if(datacheck[0] != 0x6B || datacheck[1] != 0x48 || datacheck[2] != 0x6E || datacheck[8] != datacheck[12] || datacheck[9] != datacheck[13] || datacheck[10] != datacheck[14] || datacheck[11] != datacheck[15]) {
		genClose(fd);
		return -11; // VCD has invalid header data (CUE2POPS identifier + end LBA)
	}
	genClose(fd);

	/* Making the POPStarter launch argument */

	if(mode == FALSE) { // Let's make sure we aren't in the LaunchKey config menu
		size = strlen(path);
		if(strncmp(tmp, "hdd0:__.POPS", 12) != 0) {
			for(fd = 8; fd < size; fd++) {
				if(path[fd] == '/' || path[fd] == '\\' || path[fd] == ':') { // Find the separator which just before the VCD file name
					path[fd] = 0x00;
					fd = 6;
					break;
				}
			}
			if(fd != 6) return -10;
		} else if(strncmp(fullpath, "mass:POPS/", 10) != 0) {
			for(fd = size; fd > 0; fd--) {
				if(path[fd] == '/' || path[fd] == '\\' || path[fd] == ':') break; // Find the last separator of the VCD path (the one just before the VCD file name)
			}
			if(fd == 0) return -10; // Separator not found
		}

		strcpy(exec_args[0], "uLE:"); // First thing we need to put in the arg is some fake device identifier, followed by a separator
		if(!strncmp(tmp, "hdd0:", 5)) {
			strcpy(exec_args[0] + 4, path + fd); // For the new POPStarter 13 launch type, we use the VCD name, without a prefix
		} else if(!strncmp(fullpath, "mass:POPS/", 10)) {
			strcpy(exec_args[0] + 4, "XX."); // The prefix for loading VCDs from mass, it's XX.
			strcpy(exec_args[0] + 7, fullpath + 10);
		}
		size = strlen(exec_args[0]);
		if(strncmp(tmp, "hdd0:__.POPS", 12) != 0 && strncmp(fullpath, "mass:POPS/", 10) != 0) size += 4;
		strcpy(exec_args[0] + size-4, ".elf"); // And to complete the POPStarter launch argument, we replace the VCD extension with an elf extension

		/* Loading POPStarter */
		if(!strncmp(tmp, "hdd0:", 5)) {
			strcpy(tmp, "hdd0:__common");
			strcpy(fullpath, "pfs0:/POPS/");
			if((ret = mountParty(tmp)) < 0) return -14; // Cannot mount the __common partition
			fullpath[3] += ret;
			if(0 > fileXioGetStat(fullpath, &stat)) return -15; // POPS directory wasn't found
			strcpy(fullpath + 11, "POPSTARTER.ELF");
		} else if(!strncmp(fullpath, "mass:POPS/", 10)) {
			strcpy(fullpath, "mass:POPS/POPSTARTER.ELF");
		}
		if((fd = genOpen(fullpath, O_RDONLY)) < 0) return -15; // Cannot open hdd0:__common:pfs%d:/POPS/POPSTARTER.ELF / mass:POPS/POPSTARTER.ELF for loading

		size = genLseek(fd, 0, SEEK_END);
		genLseek(fd, 0x400, SEEK_SET); // Fixed ELF header size (POPStarter Revision 13 attributes)
		genRead(fd, (void *)0x00100000, size - 0x400); // Fixed load address (POPStarter Revision 13 attributes)
		genClose(fd);

		if(_lw(0x00100000) != 0x0804000C) return -17; // Not a POPStarter ELF

		// POPStarter is loaded. We can poke its config table if needed.
		// _sb(0x04, 0x00100010);

		unmountAll();
		if(!strncmp(exec_args[0] + 4, "XX.", 3) && hdd_booted == 1) fileXioDevctl("hdd0:", 0x00004806, NULL, 0, NULL, 0); // We're launching it in USB mode. PowerOff the HDD.

		CleanUp();

		fioExit();
		SifInitRpc(0);
		SifExitRpc();
		FlushCache(0);
		FlushCache(2);

		ExecPS2((void *)0x00100000, 0, 1, exec_args);
	}

	return 1; // Tell the LaunchKey config menu we have a valid VCD
}
